!****************************************************************************
!
! File : exchange_corr.f90
!   by : Alan Tackett
!   on : 9/19/95
!  for : PAW project
!
!  NOTE : Original code was written by Dr. N.A.W. Holzwarth
!
!  Contains a routine for calculating the exchange-correlation.  Uses 
!  Kohn-Sham exchange and adds relvativistic effects if Relativistic=.TRUE.
!
!****************************************************************************

Module Exchange_Corr

             !*** Define Exchange correlation types supported ****
Integer, PARAMETER :: XC_PERDEW = 1
Integer, PARAMETER :: XC_VWN    = 2
Integer, PARAMETER :: XC_WIGNER = 3


Real, PRIVATE, PARAMETER :: BALPHA = 137.03602    !** Relativistic PARAMETER

                  !**** VWN PRIVATE, PARAMETERs ****
Real, PRIVATE, PARAMETER :: t1 = -0.10498
Real, PRIVATE, PARAMETER :: a = 0.0621814
Real, PRIVATE, PARAMETER :: b = 3.72744
Real, PRIVATE, PARAMETER :: c = 12.9352
Real, PRIVATE, PARAMETER :: t2 = t1*t1 + b*t1 + c
Real, PRIVATE, PARAMETER :: q = 6.152029831
Real, PRIVATE, PARAMETER :: c1 = 2*b/q
Real, PRIVATE, PARAMETER :: c2 = b*t1/t2
Real, PRIVATE, PARAMETER :: c3 = 2*(b+2*t1)/q 

                  !**** Perdew-Wang PRIVATE, PARAMETERs ****
Real, PRIVATE, PARAMETER :: pa = 0.031091
Real, PRIVATE, PARAMETER :: palp1 = 0.21370
Real, PRIVATE, PARAMETER :: pb1 = 7.5957
Real, PRIVATE, PARAMETER :: pb2 = 3.5876
Real, PRIVATE, PARAMETER :: pb3 = 1.6382
Real, PRIVATE, PARAMETER :: pb4 = 0.492940
Real, PRIVATE, PARAMETER :: tpa = 2 * pa


Contains

!****************************************************************************
!
! ExchangeCorr - Calculates the total exchange-correlation and exchange
!                Potential.
!
! Density - Electron density in units of electrons/Bohr^3
! Vxc     - Ouput exchange potential in Ryd
! Exc     - Output exchange energy/density
! Relativistic - Determines if relativistic effects are included(TRUE)
! XC_Type - Type of exchange correlation to perform:
!               XC_PERDEW - Perdew & Wang, PRB, June 15, 1992
!               XC_WVN    - Vosko, Wilk, & Nusair, Can. J.Phys. 58, 1200(1980)
!               XC_WIGNER - Wigner correlation
!                           Bylander & Kleinman, PRB vol 27, no. 6, 15 Mar 83
!                           Pickett, Comp. Phys. Reports 9, 115-118(1989)  
! 
!
!****************************************************************************

Subroutine ExchangeCorr(Density, Vxc, Exc, Relativistic, XC_Type)
  Use Mathlib

  Real,    Intent(IN)  :: Density
  Real,    Intent(OUT) :: Vxc
  Real,    Intent(OUT) :: Exc
  Logical, Intent(IN)  :: Relativistic
  Integer, Intent(IN)  :: XC_Type

  Real :: den, fa, vxnr, rsh,ff,ffp



  den = MAX(density, 1.0E-68)       !** Inelegant fix for small densities
  fa = (3*PI*PI*den)**(1.0/3.0)
  vxnr = -2*fa/PI

  Select Case (XC_Type)
    Case (XC_PERDEW)    !**** Perdew and Wang
      rs = (3/(4*pi*den))**(1.0/3.0)
      rsh=sqrt(rs)
      ff=pb1*rsh+rs*(pb2+pb3*rsh+pb4*rs)
      ffp=0.5*pb1*rsh+rs*(pb2+1.5*pb3*rsh+2*pb4*rs)
      xln=log(1.0+1.0/(tpa*ff))
      ec=-tpa*(1+palp1*rs)*xln
      vc=-tpa*(1 + 2.0/3.0*palp1*rs)*xln - &
          tpa*(1 + palp1*rs)*ffp/(3*ff*(1+tpa*ff))
      ec=2*ec
      vc=2*vc
   Case (XC_WIGNER)    !**** Wigner
      rs = (3.0/(4.0*pi*den))**(1.0/3.0)
      ec=-0.88/(7.79 + rs)
      vc=-0.88*(4.0/3.0*rs +7.79)/((7.79+rs)**2)
   Case (XC_VWN)       !**** Wilk, Vosko, and Nusair       
      x = (0.75/(den*pi))**(1.0/6.0)
      xx = x*x+b*x+c
      arg = q/(2.0*x+b)
      ec = a*(log((x*x)/xx)+c1*atan(arg)- &
           c2*(log(((x-t1)*(x-t1))/xx)+c3*atan(arg)))
      vc = ec-1.0/3.0*a*(c/xx-(b*t1*x)/((x-t1)*xx))
  End Select

  vxc = vxnr + vc
  exnr=0.75*vxnr
  exc=exnr + ec

  if (Relativistic) then     !*** Relativistic ***
     beta=fa/balpha
     bgam=sqrt(1.d0+beta*beta)

     if (beta > 0.001) then
       rfac=(1.d0-log(beta+bgam)/(beta*bgam))   
       rrfac=rfac*(bgam/beta)
     else
       rfac=0.666666666666666666666666d0*(beta*beta)
       rrfac=0.666666666666666666666666d0*beta
     endif

     vxr=-1.5d0*vxnr*rfac
     exr=-1.5d0*exnr*(rrfac*rrfac)
     vxc=vxc+vxr
     exc=exc+exr
  endif

  Return
End Subroutine

End Module
