!******************************************************************************
!
! File : relax.f90
!   by : Alan Tackett
!   on : 11/13/98
!  for : PAW
!
! Contains the routines for performing the different smoothing or relaxation
! operators.
!
!******************************************************************************

Module relax

Use atom_data
Use mem_data
Use projectors
Use psilib
Use hamiltonian
Use projectors
Use options_data
Use work_mgr
Use debug
Use basis_lib
Use jd_misc
Use laplacian
Use metric
Implicit NONE!!!!!!!!!!!!!

Integer, PARAMETER :: CORR_APPROX = 1    !** Approx the JD correction eq.
Integer, PARAMETER :: CORR_BICGSTABL = 2 !** Solve JD corr eq using BiCGSTABL

Type Prec_List
   Complex, Pointer :: InvMat(:,:)
   Integer          :: N
End Type

Type (Prec_List), Pointer :: PrecMatrix(:)

Type PrecVec_List
   Complex, Pointer :: V(:)
End Type

Type (PrecVec_List), Pointer :: PreVec(:)

Complex, Pointer          :: Prec_Diag(:)

Complex, Pointer :: JD_Work(:,:)  !** Kq and q orthog vectors for JD solver
Complex, Pointer :: JD_Matrix(:,:) !** qKz matrix for JD Solver
Complex, Pointer :: JD_f(:)        !** Work array needed for BiCGSTAB
Integer, Pointer :: JD_Pivot(:)    !** Pivot array for jd_matrix
Integer          :: JD_N        !** Number of JD vectors

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!*RS*
!******************************************************************************

Subroutine Prec_DiagProj_RS(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)

  Integer :: atom,  i,j,k, dn
  Complex, Pointer :: Proj(:)

  atom = PLM(1,Index)

  Proj => PreVec(index)%V
  DP = 0
  i = 0;  j = 0;
  Do k=1, RS_RLE(atom)%N
     j = RS_RLE(atom)%RLE(1,k)-1
     dn = RS_RLE(atom)%RLE(2,k)
     DP(j+1:j+dn) = Diag(j+1:j+dn)*Proj(i+1:i+dn)
     
     i = i + dn
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!*TIME*
!******************************************************************************

Subroutine Prec_DiagProj_TIME(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)
 
  Integer :: Kpnt, gmax, ng, G_half, G_size
  Integer :: Prange(2)
  Integer :: i,j,k, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Proj(:), Work(:)
  Real :: c1, c3
  Complex :: Pos_M, Neg_M, c2, dot


!Complex :: PD(PLM_Max)

!PDOT = 0
!RETURN  !**REMOVE** Debug Only!!!
  
  atom = PLM(1,Index)

  Kpnt = Mem_Kpnt
  
  gmax = Gall_Size(G_LOW)

  Base_DI = PLM_Max*(Kpnt-1)  !** Calc starting disk index for Kpnt
  ng = Gpnt_Size(G_LOW) - 1

  Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)
  !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
  !G_Size = 2*G_Half - 1
  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  Work => SCFvalues%Work
  Work = 0
  Work(FFTMap_PROJ(1:G_Half)) = PreVec(index)%V(1:G_Half)
  j = Gpnt_Size(G_PROJ)+1;   k = j + G_Half-2
  Work(FFTMap_PROJ(j:k)) = PreVec(index)%V(G_Half+1:)

  Call PerformFFT(FFT_TO_R, G_PROJ, Work)

  Work = Diag*Work

  DP = 0
  Call RS_toG(Work, DP)

!DP(1:G_Half) = PreVec(index)%V(1:G_Half)  !**QWERT
!DP(j:k) = PreVec(index)%V(G_Half+1:)     !QWERT
  Return
End Subroutine

!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!*MEM*
!******************************************************************************

Subroutine Prec_DiagProj_MEM(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)

  STOP

  Return
End Subroutine



!******************************************************************************
!
!  Prec_DiagProj - Calculates the Diagonal matrix in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     Diag  - Diag vector in R-space
!     DP    - Result of diag*Proj
!******************************************************************************

Subroutine Prec_DiagProj(index, Diag, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN)  :: Diag(:)
  Complex, intent(OUT) :: DP(:)
 
  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Call Prec_diagProj_TIME(index, Diag, DP)
     Case (MIN_MEMORY)
        Call Prec_DiagProj_MEM(index, Diag, DP)
     Case (PROJ_RS)
        Call Prec_DiagProj_RS(Index, Diag, DP)
   End Select

  Return
End subroutine

!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!*RS*
!******************************************************************************

Complex Function Prec_ProjDiagProj_RS(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)

  Integer :: atom,  i,j,k, dn, Base_DI
  Complex, Pointer :: Proj(:)
  Complex :: dot

  atom = PLM(1,Index)

  Base_DI = RS_PLMMax(atom)*(Mem_Kpnt-1)  !** Calc starting DI for Kpnt
  i = index - PLM_AtomRange(1,atom)+1
  Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + i, Proj)

  dot = 0
  i = 0;  j = 0;
  Do k=1, RS_RLE(atom)%N
     j = RS_RLE(atom)%RLE(1,k)-1
     dn = RS_RLE(atom)%RLE(2,k)
     dot = dot + SUM(Proj(i+1:i+dn)*DP(j+1:j+dn))
     
     i = i + dn
  End Do

  Prec_ProjDiagProj_RS = dot/FFT_Grid(4,G_PROJ)

  Return
end function

!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!*MEM*
!******************************************************************************

Complex Function Prec_ProjDiagProj_MEM(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)


  STOP

  Prec_ProjDiagProj_MEM = 0

  Return
end Function

!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!*TIME*
!******************************************************************************

Complex Function Prec_ProjDiagProj_TIME(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)

  Integer :: Kpnt, gmax, ng, G_half, G_size
  Integer :: Prange(2)
  Integer :: i,j,k, Base_DI,  atom, L, m, am, n, RadP, PLM_Index
  Complex, Pointer :: Proj(:), Work(:)
  Real :: c1, c3, dv
  Complex :: Pos_M, Neg_M, c2, dot
  Complex, EXTERNAL :: ZDOTC

  DV = 1.0/(1.0*FFT_Grid(4,G_PROJ))

!Complex :: PD(PLM_Max)

!PDOT = 0
!RETURN  !**REMOVE** Debug Only!!!
  
  atom = PLM(1,Index)

  Kpnt = Mem_Kpnt
  
  gmax = Gall_Size(G_LOW)

  Base_DI = PLM_Max*(Kpnt-1)  !** Calc starting disk index for Kpnt
  ng = Gpnt_Size(G_LOW) - 1

  Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)
  !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
  !G_Size = 2*G_Half - 1

  G_Half = Gpnt_Size(G_LOW)
  G_Size = Gall_Size(G_LOW)

  RadP = RS_PLM(atom,1,n)
  n = Index - PLM_AtomRange(1,atom) + 1
  Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj)

  
!  Allocate(Work(FFT_grid(4,G_PROJ)))
!  Work = 0
!  Work(FFTMap_PROJ(1:G_Half)) = Proj(1:G_Half)
!  j = Gpnt_Size(G_PROJ)+1;   k = j + G_Half-2
!  Work(FFTMap_PROJ(j:k)) = Proj(G_Half+1:)

!  Call PerformFFT(FFT_TO_R, G_PROJ, Work)

!  Prec_ProjDiagProj_TIME = DV*SUM(Work*DP)

!  DeAllocate(Work)

  Dot = SUM(Proj(1:G_Half)*DP(1:G_HAlf))
  j = Gpnt_Size(G_PROJ)+1;   k = j + G_Half-2
  Prec_ProjDiagProj_Time = dot + SUM(Proj(G_Half+1:)*DP(j:k))

  Return
end Function


!******************************************************************************
!
!  Prec_ProjDiagProj - Calculates the Proj*D*Proj in R-space times the 
!          specified projector.
!
!     index - PLM index of the projector
!     DP    - Diag*Proj
!******************************************************************************

Complex Function Prec_ProjDiagProj(index, DP)
  Integer, Intent(IN)  :: index
  Complex, intent(IN) :: DP(:)
 
  Select Case (Proj_Mode)
     Case (MIN_TIME)
        Prec_ProjDiagProj = Prec_ProjdiagProj_TIME(index, DP)
     Case (MIN_MEMORY)
        Prec_ProjDiagProj = Prec_ProjDiagProj_MEM(index, DP)
     Case (PROJ_RS)
        Prec_ProjDiagProj = Prec_ProjDiagProj_RS(Index, DP)
   End Select

  Return
End Function

!******************************************************************************
!
!  MakePrecondVects - Makes the prconditioned vectors
!
!******************************************************************************

Subroutine MakePrecondVects(Energy)
  Real, Intent(IN) :: Energy

  Integer :: nili,li,mi, njlj,lj,mj, i, j, atom, Base_DI,  n
  Complex, Pointer :: Proj1(:), Proj2(:), V(:)
  Complex :: term
  Real    :: dot


  Do atom=1, Specific_Atoms
     Base_DI =  RS_PLMMax(atom)*(Mem_Kpnt-1)

     Do i=PLM_AtomRange(1,atom), PLM_AtomRange(2,Atom)
        nili = PLM(2,i); li = PLM(3,i);  mi = PLM(4,i);
!        n = i - PLM_AtomRange(1,atom)+1
!        Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj1)

        V => PreVec(i)%V
        V = 0
        Do j=PLM_AtomRange(1,atom), PLM_AtomRange(2,Atom)
           njlj = PLM(2,j); lj = PLM(3,j);  mj = PLM(4,j);
           term = DijMatrix(atom,njlj,nili,lj,li,mj,mi) -  &
                Energy*OijMatrix(atom,njlj,nili,lj,li,mj,mi)
           term = CONJG(term)
           dot = term*CONJG(term)
           if (abs(dot)>1E-20) then
!write(Log_Unit,*) 'MPV: ij=',i,j,' * term=',term
              n = j - PLM_AtomRange(1,atom)+1
              If (Proj_Mode == PROJ_RS) then
                 Call LRU_GetRec(LRU_RS_Proj(atom),  Base_DI + n, Proj2)
              else
                 Call LRU_GetRec(LRU_Proj(atom),  Base_DI + n, Proj2)
              End If
              V = V + CONJG(term)*Proj2
           End if
        End do

        V = CONJG(V)
     End do
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  MakePrecondMatrix - Makes the preconditioner matrix for the Jacobi
!     davidson solver
!
!    Energy - Shift energy for preconditioner
! 
!******************************************************************************

Subroutine MakePrecondMatrix(Energy)
  Real, Intent(IN) :: Energy
  
  Complex, Pointer :: Matrix(:,:), Work(:), Temp1(:,:), Temp2(:,:)
  Complex          :: PDOT(PLM_Max), term
  Real             :: del2
  Integer          :: atom, i, j,k, N,  Pivot(PLM_Max)
  Integer          :: nili,li,mi, njlj,lj,mj

!RETURN !**FIXME**

  Shift_energy = energy
  !** Calculate the diagonal component of DEL2 
  del2 = 0
  Do i=1, 3
     del2 = del2 + IGij(i,i)*FFT_Grid(i,G_PROJ)**2
  End do
!  del2 = -2.0*del2
  del2 = -14350/5040.0*del2
!del2 = 0 !QWERT
  jd_diag_del2 = del2

  Call Start_Timer(Timer(JDQZSetupPrec_Timer))             


  If (Atomic_Mode /= SIM_SEPM) then
     If (Proj_Mode == PROJ_RS) then
        Work => SCFvalues%Work
     else
        Call GetBuffer( Work)
     End If

     Call MakePrecondVects(Energy)
  End If

Write(log_Unit,*) 'MakePrecMatrix: Vo=',SCFvalues%Ve_Fourier(1)

  !*** Now construct the diagonal and invert it ***
  Prec_Diag = -del2 + V_Weight*SCFvalues%Ve - Energy
  Do i=1, FFT_Grid(4,G_PROJ)
     del2 = CONJG(Prec_Diag(i))*Prec_Diag(i)
     if (abs(del2) > 1E-20) then 
        prec_diag(i) = 1.0/Prec_Diag(i)
     else
        Write(Log_Unit,*) 'MakePrecMatrix: ZERO Diag! i=',i
     End IF
  End do

  If (Atomic_Mode /= SIM_SEPM) then
     !*** Now form the matrix ***
     Do atom=1, Specific_Atoms
        N = PrecMatrix(atom)%N
        Matrix => PrecMatrix(atom)%InvMat
        
        k = PLM_AtomRange(1,atom)-1
        Do i=PLM_AtomRange(1,atom), PLM_AtomRange(2,atom)
           Call Prec_DiagProj(i, Prec_Diag, Work)
           nili = PLM(2,i); li = PLM(3,i); mi = PLM(4,i)
           Do j=PLM_AtomRange(1,atom), PLM_AtomRange(2,atom)
              njlj = PLM(2,j); lj = PLM(3,j); mi = PLM(4,j)
              Matrix(j-k,i-k) = Prec_ProjDiagProj(j,Work)
              
              if (j==i) then
                 Matrix(j-k,i-k) = Matrix(j-k,i-k) + 1
              End if
           End do
        End Do
        
        !*** Finally invert the matrix ***
        Call ZGETRF(N,N, Matrix(1,1), N, Pivot(1), i)
        If (i /= 0) then
           Write(Error_Unit,*) 'MakePre: Error in DGETRF.  Error=',i
           i = 0;
           j = 1.0/i;
        End If

        Call ZGETRI(N, Matrix(1,1), N, Pivot(1), Work, FFT_Grid(4,G_PROJ), i)
        If (i /= 0) then
           Write(Error_Unit,*) 'MakePre: Error in DGETRI.  Error=',i
           i = 0;
           j = 1/i;
        End If
     End do

     Call FreeBuffer(Work)
  End If

  Call Stop_Timer(Timer(JDQZSetupPrec_Timer))             

  Return
end Subroutine

!******************************************************************************
!
!  JD_Precond - Applies the approximate Jacobi-Davidson inverse precond
!
!******************************************************************************

Subroutine JD_Precond(Precond, Psi, Kpnt)
  Complex, Intent(OUT) :: PreCond(:)
  Complex, Intent(IN)  :: Psi(:)
  Integer, Intent(IN)  :: Kpnt

  Logical :: D2_separate
  Integer :: i, j,k,index,dn, G_half, G_Size, atom
  Complex, Pointer :: work(:), Work2(:), Proj(:), D2x(:)
  Complex :: PDOT(PLM_Max,2), Vpnt, den
  Real    :: c1, Kvec(3), g2, g(3)
  Integer :: gi, N

  D2_Separate = .FALSE. !** ALSO Make Sure in MakePreCondMatrix del2=0!!!

  Work => SCFvalues%Work

  If (Atomic_Mode == SIM_SEPM) then
     Work = 0
     Call RS_toR(Psi, Work) 

     Work = Prec_diag*Work
     PreCond = 0
     Call RS_toG(Work, PreCond)
  End If

  If (D2_Separate) then
     Call GetBuffer( D2x)

     D2x = 0
     D2x(1) = Psi(1)

     c1 = -1

     Kvec = BZ%Ku(1:3,Mem_Kpnt)

     N = Gpnt_Size(G_LOW)-1
     
     Do gi=2, Gpnt_Size(G_LOW)
        G = Kvec + Gpnt(1:3,gi)
        G2 = DOT_PRODUCT(G,G)
        D2x(gi) = c1*Psi(gi)/G2
     End Do
     Do gi=2, Gpnt_Size(G_LOW)
        G = Kvec - Gpnt(1:3,gi)
        G2 = DOT_PRODUCT(G,G)
        D2x(gi+N) = c1*Psi(gi+N)/G2
     End Do

     Work = 0
     Call RS_toR(D2x, Work) 
  else
     Work = 0
     Call RS_toR(Psi, Work) 
  End If
 
  PreCond = 0

!  Work = 0
!  Call RS_toR(Psi, Work) 

  Work = Prec_Diag*Work

  !*** Calculate Dx ****
  If (Proj_mode == PROJ_RS) then
       Call CalcProjProducts_RS(Work, PDOT(:,1))   
  else
     Call RS_toG(Work, PreCond)
     Call CalcProjProducts(PreCond, PDOT(:,1))
  End If
!PreCond = Psi !QWERT
!RETURN


  !** Multiply by matrix ***
  Do atom=1, Specific_atoms
     i=PLM_AtomRange(1,atom); j=PLM_AtomRange(2,atom)

     PDOT(i:j,2) = MatMul(PrecMatrix(atom)%InvMat, PDOT(i:j,1))
  End Do


  If (Proj_Mode ==PROJ_RS) then
     Allocate(Work2(FFT_Grid(4,G_PROJ)))
     Work2 = 0
     Do index=1, PLM_Max
        atom = PLM(1,index)
        Proj => PreVec(index)%V
        i = 0;  j = 0;
        Do k=1, RS_RLE(atom)%N
           j = RS_RLE(atom)%RLE(1,k)-1
           dn = RS_RLE(atom)%RLE(2,k)
           Work2(j+1:j+dn) = Work2(j+1:j+dn) - Proj(i+1:i+dn)*PDOT(index,2)
           i = i + dn
        End Do        
     End do

     Work = Work + Prec_Diag*Work2
     DeAllocate(Work2)
  else
     Call GetBuffer( Work2)
     Work2 = Psi
     Do i=1, PLM_Max
        atom = PLM(1,i)
        !G_Half = Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size
        !G_Size = 2*G_Half - 1
        G_Half = Gpnt_Size(G_LOW)
        G_Size = Gall_Size(G_LOW)
        
        Work2(1:G_half) = Work2(1:G_Half) - PreVec(i)%V(1:G_Half)*PDOT(i,2)
        j = Gpnt_Size(G_PROJ) + 1; k = j + G_Half - 2
        Work2(j:k) = Work2(j:k) - PreVec(i)%V(G_Half+1:)*PDOT(i,2)     
     End Do
     
     
     Work = 0
     Call RS_toR(Work2, Work) 
     Work = Prec_Diag*Work
     Call FreeBuffer(Work2)
  End If
  
  PreCond = 0
  Call RS_toG(Work, PreCond)


!D2_Separate = .TRUE.

  If (D2_Separate) then
     c1 = -1

     Kvec = BZ%Ku(1:3,Mem_Kpnt)

     N = Gpnt_Size(G_LOW)-1
     
     Do gi=2, Gpnt_Size(G_LOW)
        G = Kvec + Gpnt(1:3,gi)
        G2 = DOT_PRODUCT(G,G)
        Precond(gi) = D2x(gi) - c1*Precond(gi)/G2
!        Precond(gi) = Precond(gi) + c1*Psi(gi)/G2
     End Do
     Do gi=2, Gpnt_Size(G_LOW)
        G = Kvec - Gpnt(1:3,gi)
        G2 = DOT_PRODUCT(G,G)
        Precond(gi+N) = D2x(gi+N) - c1*Precond(gi+N)/G2
!        Precond(gi+N) = Precond(gi+N) + c1*Psi(gi+N)/G2
     End Do

     Call FreeBuffer(D2x)
  End If

!PreCond = Work2 !QWERT


  Return
End Subroutine

!******************************************************************************
!
! GS_Test_Precond - Tests the precond code in G-space
!
!******************************************************************************

Subroutine GS_Test_Precond
  Complex, Pointer :: X(:), Vinv(:), V(:), Hx(:), Ox(:)
  Complex :: dot, dot2, PDOT(PLM_Max)
  Integer :: i,j, Kpnt,  N
  Real    :: r, dv
  Logical :: PDOT_Stored
Complex, Pointer :: Amat(:,:), Ainvmat(:,:)
Write(Log_Unit,*) 'GS_Test_Precond: Start!!!!!!!!!!!!!!!!!!! PLM_Max=',PLM_Max

!  N = FFT_Grid(4,G_PROJ)
  N = Gall_Size(G_LOW)
  Allocate(X(N), Vinv(N), V(N), Hx(N), Ox(N))
Allocate(AinvMat(N,N))
Mem_Kpnt = 1
dv = 1.0/FFT_Grid(4,G_PROJ)


!Omat = 0
!Do i=1,n
!   Omat(i,i) = Omat(i,i) + 1
!   Do j=1, n
!      Omat(i,j) = Omat(i,j) + V(i)*CONJG(V(j))
!   End Do
!end Do
!Call RS_Oinverse(V, 1)

  Call MakePrecondMatrix(JDQZ_Energy)

  Do Kpnt=1, NumKpnts  
     Write(Log_Unit,*) 'GS_Test_Oinverse: checking Kpnt=',kpnt
     Mem_Kpnt = Kpnt

     Do i=1, N
        X = 0
        X(i) = 1
        PDOT = 0
        PDOT_Stored = .FALSE.
        Call CalcHxandOx( x, Hx, Ox, &
             PDOT, Kpnt, PDOT_Stored)
        X = Hx - JDQZ_Energy*Ox
!Write(Log_Unit,*) 'i=',i, ' * DOT(Ox)=',DOT_PRODUCT(Ox,Ox), ' * DOT(Hx)=',Dot_Product(Hx,Hx)
!Do j=1, N
!  dot = O(j) - Omat(j,i)
!  r = CONJG(dot)*dot
!  if (ABS(r) > 1E-9) then
!     Write(Log_Unit,*) 'Omat: ij=',i,j, ' * Oj=',O(j), ' * Oji=',Omat(j,i), ' * ERR=',r
!  End if
!End Do
V = 0
V(i) = 1
Call JD_Precond(AinvMat(:,i), V, Kpnt)
AinvMat(:,i) = X
        Do j=1, i
           V = 0
           V(j) = 1
           
           Call JD_Precond(Vinv, V, Kpnt)
           dot = DOT_PRODUCT(X,Vinv)
           
           r = CONJG(dot)*dot
           r = SQRT(r)
           If (i==j) then
              if (ABS(r-1) > 1E-9) then
                 write(Log_Unit,*) 'ij=',i,j,' * dot=',dot
              end If
           else
              if (ABS(r) > 1E-9) then
                 write(Log_Unit,*) 'ij=',i,j,' * dot=',dot
              end If
           End If
        End Do
     End Do
Write(Log_Unit,*) 'AinvMat!!!!!!!!!!!!!!!!!!!!!!!!!!!'
!Do i=1, N
!   Write(Log_Unit,*) 'i=',i, ' * ',AinvMAt(i,:)
!End Do
Do i=1, N
   Do j=1,i-1
      dot = AinvMAt(i,j) - CONJG(AinvMat(j,i))
      r = CONJG(dot)*dot
      If (ABS(r) > 1E-20) then
         Write(Log_Unit,*) ' ij=',i,j,' * ',ainvmat(i,j), ainvmat(j,i)
      End if
   End Do
End Do
  End do


Write(Log_Unit,*) 'GS_Test: Completed!'
  DeAllocate(X, Vinv, Hx, Ox)

  STOP
!  Call FreeBuffer(O)
!  Call FreeBuffer(Oinv)
!  Call FreeBuffer(V)

  Return
End Subroutine

!******************************************************************************
!
! SetupJDCorr - Sets up the JD vectors and matrices for a call to
!         ZCGSTAB which attempts to solve the JD correction equation.
!         This routine assumes that the preconditioner has already
!         been setup.
!
!   PsiList - List of Vectors to orthog to in the JDQZ solver
!
!******************************************************************************

Subroutine SetupJDCorr(PsiList)
  Integer, Intent(IN) :: PsiList(:)

  Complex, PARAMETER :: one  = (1,0)
  Complex, PARAMETER :: zero = (0,0)

  Integer :: i, j,  N
  complex :: dot

  Call Start_Timer(Timer(JDQZSetupCorr_Timer))             


  !** Free old mem if needed **
  If (JD_N>0) DeAllocate(JD_Work, JD_Matrix, JD_Pivot, JD_f)

  JD_N = SUM(PsiList)  !** Determine how many vectors there are

!Write(Log_Unit,*) 'SJDC: JD_N=',jd_n

  n = Gall_Size(G_LOW)
  Allocate(JD_Work(n,2*JD_N+2*2+6), JD_Matrix(JD_N, JD_N), &
       JD_Pivot(JD_N), JD_f(JD_N), STAT=i) 
  If (i /= 0) then
     Write(Error_Unit,*) 'SetupJDCorr: Allocate Error! Error=',i
     dot = 0
     dot = 1.0/dot
     STOP
  End if

  i = 0
  Do j=1, Mem_MapSize
     If (PsiList(j) == MH_toProcess) then
        i = i + 1
!Write(Log_Unit,*) 'SJDC: i=',i, ' * j=',j
        Call GetPsi_fromBuffer( j, JD_Work(:,i))
!Write(Log_Unit,*) 'SJDC:  PsiDOT=',DOT_PRODUCT(JD_Work(:,i),JD_Work(:,i))
        Call zmgs(n, i-1, JD_work(1,1), JD_work(1,i), 1)  !** Orthog to others
!Write(Log_Unit,*) 'SJDC:     ODOT=',DOT_PRODUCT(JD_Work(:,i),JD_Work(:,i))
        JD_Work(:,JD_N+i) = JD_Work(:,i)   !** Make the precond version
        Call Precon(n, JD_Work(1,JD_N+i))
!Write(Log_Unit,*) 'SJDC:   KDOT=',DOT_PRODUCT(JD_Work(:,JD_N+i),JD_Work(:,JD_N+i))
        Call zmgs(n, i-1, JD_Work(1,JD_N+1), JD_work(1,JD_N+i), 1)
!Write(Log_Unit,*) 'SJDC:   OKDOT=',DOT_PRODUCT(JD_Work(:,JD_N+i),JD_Work(:,JD_N+i))
     End If
  End do

  !*** Make H~ matrix ***
  Call ZGEMM('C', 'N', JD_N, JD_N, n, one, JD_Work(1,JD_N+1), n, &
       JD_Work(1,1), n, zero, JD_Matrix(1,1), JD_N)

  Call ZGETRF (JD_N, JD_N, JD_Matrix(1,1), JD_N, JD_Pivot(1), i)
  If (i /= 0) then
     Write(Error_Unit,*) 'SetupJDCorr: ZGETRF: info=',i
     dot = 0
     dot = 1.0/dot
     STOP
  End if

  Call Stop_Timer(Timer(JDQZSetupCorr_Timer))             

  Return
End subroutine


!******************************************************************************
!
!  KE_PreCond - Constructs the KE preconditioner
!
!  Diag - KE preconditioner returned
!  KE   - Ke for band
!  BlochK - K-point
!
!******************************************************************************

Subroutine KE_Precond(Diag, KE, BlochK)
  Complex, Intent(OUT) :: Diag(:)
  Real,    Intent(IN)  :: KE
  Real,    Intent(IN)  :: BlochK(:)

  Integer :: gi, N
  Real    :: G(3), G2, x, poly

  N = Gpnt_Size(G_LOW) - 1

  Do gi=1, Gpnt_Size(G_LOW)          
     G = BlochK + Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     
     x = G2/KE
     poly = 8*x**3 + 12*x**2 + 18*x + 27
     Diag(gi) = poly/(poly + 16*x**4)
  End Do
        
  Do gi=2, Gpnt_Size(G_LOW)
     G = BlochK - Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     
     x = G2/KE
     poly = 8*x**3 + 12*x**2 + 18*x + 27
     Diag(gi+N) = poly/(poly + 16*x**4)
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  JD_CalcCorr - Calculates a correction to the Psi's  using a 
!       Block-Davidson step
!
!***NEW Process a single vec***
!******************************************************************************

Subroutine JD_CalcCorr( Psi, Hx, Ox, PsiIndex, Kpnt, CorrType)
  Complex,          Intent(IN) :: Psi(:)
  Complex,          Intent(IN) :: Hx(:)
  Complex,          Intent(IN) :: Ox(:)
  Integer,          Intent(IN) :: PsiIndex
  Integer,          Intent(IN) :: Kpnt
  integer,          Intent(IN) :: CorrType

  Complex, Pointer :: Diag(:), Work(:), Work2(:), OWork2(:)
  Real :: Energy, BlochK(3), G(3), G2, KE, deps
  Complex  :: dot, zalpha, zbeta
  Integer :: N, BG_Index, gi, L, mxmv
  Logical :: PDOT_Stored
  Complex :: PDOT(PLM_Max)


  Call Start_Timer(Timer(CalcCorr_Timer))
  N = Gpnt_Size(G_LOW) - 1

  Call GetBuffer( Diag)
  Call GetBuffer( Work)

  BlochK = BZ%Ku(:,Kpnt)

  Call GetNextFreeIndex(BG_Index)
  PsiInfo(BG_Index)%Kpnt = BG_KPNT !special index for temp vectors 
  PsiInfo(PsiIndex)%CorrIndex = BG_Index

  Energy = PSiInfo(PsiIndex)%Energy

  Select Case (CorrType)
     Case (CORR_APPROX)
        KE = CalcKineticenergy(BZ%Ku(1:3,Kpnt), Psi)
        Call KE_Precond(Diag, KE, BlochK)
        Work = Diag*(Hx - Energy*Ox)


!*3*-------   ** Working Scheme **
        If (Eigen_Mode == EIGEN_DAVIDSON) then
           Diag = Diag*Psi
           dot = Basis_DotProd( Ox, Work, .FALSE.)
           dot = dot / Basis_DotProd( Ox, Diag, .FALSE.)
           Work = dot*Diag - Work
        End If
!--------
     Case (CORR_BiCGSTABL)
        L = 2
        Diag = Hx - Energy*Ox
        zalpha = Energy; zbeta = 1.0;
        mxmv = JDQZ_MaxMV;
        deps = 1E-12
        Call ZCGSTABL(Gall_size(G_LOW), Work(1), diag(1), L, deps, &
             mxmv, zalpha, zbeta, JD_N, JD_work(1,JD_N+1), JD_work(1,1), &
             JD_Matrix(1,1), JD_N,JD_Pivot(1), &
             JD_f(1), JD_work(1,2*JD_N+1), 2*L+6)
  End Select

  Call SavePsi_toBuffer( PSiInfo(PsiIndex)%CorrIndex, Work)

 !write(Log_Unit,*)'in JD_C -- psiindex,corrindex', psiindex,&
 !psiinfo(psiindex)%corrindex,psiinfo(psiinfo(psiindex)%corrindex)%Available,&
 !psiinfo(psiinfo(psiindex)%corrindex)%membufindex
  

  Call FreeBuffer(Work)
  Call FreeBuffer(Diag)

  Call Stop_Timer(Timer(CalcCorr_Timer))  
  Return
End Subroutine


!******************************************************************************
!
!  JD_CalcCorr_Partition - Calculates a correction to the Psi's  using a 
!       Block-Davidson step for the partition energy scheme
!
!***NEW Process a single vec***
!******************************************************************************

Subroutine JD_CalcCorr_Partition( Psi, Hx, Ox, PsiIndex, Kpnt, Energy)
  Complex,          Intent(IN) :: Psi(:)
  Complex,          Intent(IN) :: Hx(:)
  Complex,          Intent(IN) :: Ox(:)
  Integer,          Intent(IN) :: PsiIndex
  Integer,          Intent(IN) :: Kpnt
  Real,             Intent(IN) :: Energy

  Complex, Pointer :: Diag(:), Work(:), W2(:), DiagPsi(:)
  Complex :: PDOT_HOiHx(PLM_Max), dot
  Logical :: PDOT_Stored
  Real :: BlochK(3), G(3), G2, x, poly, KE
!*  Complex  :: KE
  Integer :: N, BG_Index, gi


  N = Gpnt_Size(G_LOW) - 1

  Call GetBuffer( DiagPsi)
  Call GetBuffer( Diag)
  Call GetBuffer( Work)

  BlochK = BZ%Ku(:,Kpnt)

  Call GetNextFreeIndex(BG_Index)
  PsiInfo(BG_Index)%Kpnt = BG_KPNT
  PsiInfo(BG_Index)%DoSave = 1      
  PsiInfo(BG_Index)%Available=Mem_Used
  PsiInfo(PsiIndex)%CorrIndex = BG_Index

  KE = CalcKineticenergy(BZ%Ku(1:3,Kpnt), Psi)

  Call KE_Precond(Diag, KE, BlochK)
   
  Work = Diag*(Hx - Energy*Ox)
  DiagPsi = Diag*Psi

  PDOT_Stored = .FALSE.
  Diag = 0
  Call OPsi( Work, Diag, PDOT_HoiHx, Kpnt, PDOT_Stored)
  Work = Diag

  PDOT_Stored = .FALSE.
  Diag = 0
  Call OPsi( DiagPsi, Diag, PDOT_HoiHx, Kpnt, PDOT_Stored)
  DIagPsi = Diag



  KE = CalcKineticenergy(BZ%Ku(1:3,Kpnt), Work)
  Call KE_Precond(Diag, KE, BlochK)
  Work = Diag*Work             !*Old Working scheme
!**  Work = Diag*Work + Energy*Work
  DiagPsi = Diag*DiagPsi

  

!*3*-------
!**  Diag = Diag*Psi
  dot = Basis_DotProd( Ox, Work, .FALSE.)
  dot = dot / Basis_DotProd( Ox, DiagPsi, .FALSE.)
  Work = dot*DiagPsi - Work
!--------

  Call SavePsi_toBuffer( PSiInfo(PsiIndex)%CorrIndex, Work)
  

  Call FreeBuffer(Work)
  Call FreeBuffer(Diag)
  Call FreeBuffer(DiagPsi)
  
  Return
End Subroutine

!******************************************************************************
!
!  AllocPrecVec - Allocatesthe precondition vectors
!
!******************************************************************************

Subroutine AllocPrecVec
  Integer :: i,j,atom

  Do i=1, PLM_Max
     atom = PLM(1,i)
     If (Proj_Mode == PROJ_RS) then
        j =LRU_RS_Proj(atom)%Data_Size
     else
        !j = 2*Atomtype_Info(Atom_List(atom)%typeIndex)%Gpnt_size - 1
        j = Gall_Size(G_LOW)
     End If

     If (JD_N>0) DeAllocate(PreVec(i)%V)
     Allocate(PreVec(i)%V(j))
  End do

  Return
End Subroutine

!******************************************************************************
!
!  Init_relax - Initializes the relax routines for use
!
!******************************************************************************

Subroutine Init_relax
  Integer :: i, j, atom

  If (Atomic_Mode == SIM_SEPM) then
     Allocate(Prec_Diag(FFT_Grid(4,G_PROJ)))
  else
     Allocate(PrecMatrix(Specific_Atoms), PreVec(PLM_Max), &
          Prec_Diag(FFT_Grid(4,G_PROJ)))

     Do i=1, Specific_Atoms
        j = PLM_AtomRange(2,i) - PLM_AtomRange(1,i)+1
        PrecMatrix(i)%N = j
        Allocate(PrecMatrix(i)%InvMat(j,j))
     End Do
     
     JD_N = -1

     Call AllocPrecVec
  End If

  Call CalcGij(xtal%Basis)

  Return
end subroutine

End Module
